/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

/******************************************************************************
 *
 *
 *	   cportdir.cpp
 *		- class definition of port 
 *		  definition nodes
 *
 ******************************************************************************
 */

#include "cportdir.h"
#include "cnode.h"




/****************************************************
	Constructor
*****************************************************/
CPortDir::CPortDir( CSymbol* name, Coord_t* aLoc, 
		    Decl_t type, CDataType* dataType ) :
	CDecl( name, aLoc, type, dataType, Or(eFLAG_DATATYPE,eFLAG_ARRAY) ) {
	decl = NULL;
}

/****************************************************
	Clone
	- return a deep copy of declaration
*****************************************************/
CDecl* CPortDir::Clone( CObstack* heap )
{
     CPortDir* clone = new(heap) CPortDir( GetSymbol(),
		GetCoord(), GetType(), NULL );
     clone->Copy( heap, *this );
     return clone;
}

/****************************************************
	Copy
	- Deep copy of declaration
*****************************************************/
void CPortDir::Copy( CObstack* heap, CPortDir& d )
{
    decl     = d.decl;
}

/****************************************************
	Dump
*****************************************************/
void	CPortDir::Dump( FILE* f )
{
	if( decl == NULL ) {
		fprintf( f, "%s: %s, defined in ", 
			declName[GetType()], GetName() );
	} else {
		fprintf( f, "%s: %s -> %s defined in ", 
			declName[GetType()], GetName(),
			decl->GetName() );
	}
	CDecl::Dump( f );
	fprintf( f, "\n" );
}	

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CPortDir::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    CDecl::PreVisit1( func, data );
    if( decl && !decl->DeclStatementCreated() ) {
	decl->PreVisit1( func, data );
    }
}

/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CPortDir::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    CDecl::PostVisit1( func, data );
    if( decl && !decl->DeclStatementCreated() ) {
	decl->PostVisit1( func, data );
    }
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CPortDir::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    CDecl::PostSubVisit1( func, data );
    if( decl && !decl->DeclStatementCreated() ) {
	decl->PostSubVisit1( func, data );
    }
}

