/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   lex.hpp
 *		- definitions for lexer 
 *
 *
 ******************************************************************************
 */

#ifndef LEX_HPP
#define LEX_HPP

#include "cnode.h"

struct Coord_t;
class CNode;
struct NodePair_t {
	CNode* h;
	CNode* t;
};

struct MacroHistory { 
    unsigned	serialNumber;
    const char* macroName;
    int		final;
    int	        initialTimestamp; 
    int	        finalTimestamp; 
};

struct MacroToken {
    int	 	null;
    char* 	text;
    CNode_pr	nodePtr;
    CAttr* 	attribute;
    CNode_pr	previousNodePtr;
    CNode*	macro;
    int	        initialTimestamp; 
    int	        finalTimestamp; 
    int		historyDepth;
    MacroHistory *history;
};

int	lex_openFile( char* filename );
char*	lex_getComments( void );
void	lex_init( void );
void	lex_startUDP( void );
void	yyerror( const char* message );
int	prog_parse( void );
void	parse_init( void );
void  clean_identifier( char* s );
MacroToken MergeToken( const MacroToken& a, const MacroToken& b );
const MacroHistory* CompleteToken( const MacroToken& macroInfo );
inline int NullMacroToken( const MacroToken& m) { return m.null; }
inline int InvalidMacroToken( const MacroToken& m) { return !m.history; }
MacroToken NullMacroToken();
MacroToken InvalidMacroToken();
CNode* Macro2Node( const MacroToken& m );
inline char* Macro2Text( MacroToken& m ) { return m.text; }
inline CAttr* Macro2Attribute( MacroToken& m ) { return m.attribute; }
void DumpMacroInfo( const MacroToken& a );

#endif // LEX_HPP
