/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

/******************************************************************************
 *
 *
 *	   main.hpp
 *		- definitions for main 
 *
 *
 ******************************************************************************
 */

#ifndef MAIN_HPP
#define MAIN_HPP

#include <stdio.h>
#include <string.h>
#include <string>
#include <list>
#include "csymtab.h"
#include "cdecl.h"

using namespace std;

/**
 * Plugins should not be using these.
 * \cond
 */
extern int inferVectors;
extern int quietFlag;
extern int ignoreVrqComments;
extern int verboseFlag;
extern int svEnable;
extern CSymtab<CDecl> cuSymtab;
/**
 * \endcond
 */

/**
 * Get string containing vrq version number.
 * \ingroup utilities
 * \return char pointer to vrq version number.
 */
const char*   VrqVersionString();
/**
 * Get plusarg from command line that matches prefix.
 * \ingroup utilities
 * \param prefix to search for.
 * \return complete plusarg text.
 */
char*	GetPlusArg( const char* prefix );
/**
 * Get list of plusargs from command line that matche prefix.
 * \ingroup utilities
 * \param prefix to search for.
 * \param args list to populate with matching plusargs.
 */
void    GetPlusArgs( const char* prefix, list<string>& args );
/**
 * Get list of paths to where plugins may reside.
 * \ingroup utilities
 * \param paths list of filesystem paths.
 */
void    GetPluginPaths( list<string>& paths );


#endif // MAIN_HPP
